from pytest import warns
from creationism.registration.factory import RegistrantFactory
from creationism.registration.utils import Text, chain_functions
from creationism.registration.warnings import DuplicateRegistrantNameWarning


class TestWarnings:
    def test_no_replace(self):
        with warns(DuplicateRegistrantNameWarning):

            class A(RegistrantFactory):
                AUTO = False
                CONVERT_NAME = lambda x: x
                REPLACE = False

            @A.register(("suba", "suba"))
            class SubA(A):
                pass

        with warns(DuplicateRegistrantNameWarning):

            class A(RegistrantFactory):
                AUTO = True
                CONVERT_NAME = lambda x: chain_functions(
                    x, Text.split_capitals_with_underscore, Text.lower
                )
                REPLACE = False

            @A.register(("sub_a",))
            class SubA(A):
                pass

    def test_replace(self):

        with warns(None) as record:

            class A(RegistrantFactory):
                AUTO = False
                CONVERT_NAME = lambda x: chain_functions(
                    x, Text.split_capitals_with_underscore, Text.lower
                )
                REPLACE = True

            @A.register(("sub_a",))
            class SubA(A):
                pass

            assert len(record) == 0

        with warns(None) as record:

            class A(RegistrantFactory):
                AUTO = False
                CONVERT_NAME = lambda x: x
                REPLACE = True

            @A.register(("suba", "suba"), replace=True)
            class SubA(A):
                pass

        assert len(record) == 0



import os
from creationism.configuration.config import Configuration
from creationism.configuration.config import open_config
from creationism.configuration.extensions import ConfigurationFileExtension, JsonConfigurationFileExtension
from creationism.mode import Mode, DefaultMode
from pathlib import Path

from creationism.utils import create_name_from_class_name


class TestConfig:
    def test_create_config_dict(self):
        config = {"a": 1, "b": "b", "c": [1, 2, 3]}
        config_dict = Configuration(name='', config_value=config, modes=('default',))
        assert isinstance(config_dict, Configuration)

    def test_create_config_dict_with_replace_is_true_list(self):
        config = {"a": 1, "b": "b", "c@replace=true": [1, 2, 3]}
        config_dict = Configuration(name='', config_value=config, modes=('default',))
        assert config_dict["c"].replace is True

    def test_create_config_dict_with_replace_is_false_list(self):
        config = {"a": 1, "b": "b", "c@replace=false": [1, 2, 3]}
        config_dict = Configuration(name='', config_value=config, modes=('default',))
        assert config_dict["c"].replace is False

    def test_create_config_dict_with_replace_is_true_dict(self):
        config = {"a": 1, "b": "b", "c@replace=true": {"c2": "hello"}}
        config_dict = Configuration(name='', config_value=config, modes=('default',))
        assert config_dict["c"].replace is True

    def test_create_config_dict_with_replace_is_true_dict(self):
        config = {"a": 1, "b": "b", "c@replace=false": {"c2": "hello"}}
        config_dict = Configuration(name='', config_value=config, modes=('default',))
        assert config_dict["c"].replace is False

    def test_create_config_dict_with_yaml_reference(self):
        config = {"a": 1, "b": "b", "c": str(Path(__file__).parent / "test.yml")}
        config_dict = Configuration(name='', config_value=config, modes=('default',)).cast()
        assert config_dict["c"]["name"] == [1, 2, 3]

    def test_create_config_dict_with_yaml_reference_replace(self):
        config = {"a": 1, "b": "b", "c": str(Path(__file__).parent / "test.yml")}
        config_dict = Configuration(name='', config_value=config, modes=('default',))
        assert config_dict["c"]["name"].replace is False

    def test_merge_replace_dict_true(self):
        config = {"a": 1, "b": "b", "c": {"k": {"n": 4}}}
        config_dict = Configuration(name='1', config_value=config, modes=('default',))

        config2 = {"a": 1, "b": "b", "c@replace=true": {"k": {"l": 5}}}
        config_dict.merge(config2)
        assert config_dict["c"]["k"]["l"].cast() == 5
        assert "n" not in config_dict["c"]["k"]

    def test_merge_replace_dict_false(self):
        config = {"a": 1, "b": "b", "c": str(Path(__file__).parent / "test.yml")}
        config_dict = Configuration(name='', config_value=config, modes=('default',))
        assert config_dict["c"]["name"].replace is False

    def test_merge_replace_list_true(self):
        config = {"a": 1, "b": "b", "c": str(Path(__file__).parent / "test.yml")}
        config_dict = Configuration(name='', config_value=config, modes=('default',))
        assert config_dict["c"]["name"].replace is False

    def test_merge_replace_list_false(self):
        config = {"a": 1, "b": "b", "c": str(Path(__file__).parent / "test.yml")}
        config_dict = Configuration(name='', config_value=config, modes=('default',))
        assert config_dict["c"]["name"].replace is False

    def test_has_extension(self):
        user_config = Path(__file__).parent / "testobject.json"
        assert True == ConfigurationFileExtension.has_extension(user_config, JsonConfigurationFileExtension)

    def test_is_extension(self):
        assert True == ConfigurationFileExtension.is_extension('.json', JsonConfigurationFileExtension)

    def test_create_name_from_class(self):
        assert 'Json Configuration File' == create_name_from_class_name('JsonConfigurationFileExtension')

    def test_build_config(self):
        user_config = str(Path(__file__).parent / "testobject.json")
        config_value = open_config(user_config, search_paths=('./', ))
        configuration = Configuration(name='configuration', config_value=config_value, modes=('default',))
        builds = configuration.build_instances()
        assert builds['configuration']['default']['mode'] is Mode.create('default')

    def test_build_config_class(self):
        _DEFAULT_MODES = ("test",)
        class TestConfiguration(Configuration):
            NAME = "testconfiguration"
            PRESETS_FOLDER = Path(__file__).absolute().parent / "presets"
            CONFIG_PATH = Path(__file__).absolute().parent / os.path.join(
                "config_files", "config.yml"
            )

            SEARCH_PATHS = ("", Path(__file__).absolute().parent / "config_files")

            def __init__(self, modes=_DEFAULT_MODES, search_paths=()):

                search_paths_ = self.__class__.SEARCH_PATHS + search_paths
                config_value = open_config(
                    TestConfiguration.CONFIG_PATH, search_paths=search_paths_
                )

                super().__init__(
                    name=TestConfiguration.NAME,
                    modes=modes,
                    config_value=config_value,
                    search_paths=search_paths,
                )
        


        user_config = str(Path(__file__).parent / "testobject.json")
        config_value = open_config(user_config, search_paths=('./', ))
        user_config_dict = open_config(user_config, search_paths=('./', ))
        configuration = Configuration(name='configuration', config_value=config_value, modes=('default',))
        builds_external = TestConfiguration.build(user_config=user_config_dict, modes=('default',), external_configurations=(configuration, ))

        user_config = str(Path(__file__).parent / "testobject2.yml")
        build_builder = TestConfiguration.build(user_config=user_config_dict, modes=('test',), build_instances=False)
        builds = TestConfiguration.build(user_config=user_config, modes=('test',), presets=('aisb', ))
        builds = TestConfiguration.build(user_config=user_config_dict, modes=('test',), search_paths=('./', ))


        ref = TestConfiguration.build(user_config=user_config, modes=('test',), build_key='ref')

        assert isinstance(ref['test'], DefaultMode)



from creationism.configuration.extensions import ConfigurationFileExtension, JsonConfigurationFileExtension
from creationism.registration.factory import RegistrantFactory
from creationism.registration.utils import Text, chain_functions


class TestRegistrar:
    def test_not_auto_register(self):
        class A(RegistrantFactory):
            AUTO = False

        class SubA(A):
            pass

        assert A._REGISTER is None

    def test_auto_register(self):
        class A(RegistrantFactory):
            AUTO = True
            CONVERT_NAME = lambda x: chain_functions(
                x, Text.split_capitals_with_underscore, Text.lower
            )

        class SubA(A):
            pass

        assert A._REGISTER is not None
        assert "sub_a" in A._REGISTER
        assert SubA in A._REGISTER.values()

    def test_decorator_register(self):
        class A(RegistrantFactory):
            AUTO = False

        @A.register(("suba",))
        class SubA(A):
            pass

        assert A._REGISTER is not None
        assert "suba" in A._REGISTER
        assert SubA in A._REGISTER.values()



    def test_decorator_register_recursive(self):
        class A(RegistrantFactory):
            AUTO = False

        @A.register(("suba",), recursive=True)
        class SubA(A):
            pass

        assert A._REGISTER is not None
        assert "suba" in A._REGISTER
        assert SubA in A._REGISTER.values()


    def test_name(self):
        class A(RegistrantFactory):
            AUTO = True

        class SubA(A):
            pass

        sub_a = SubA()
        assert sub_a.name == 'SubA'



    def test_convert_name_name(self):
        class A(RegistrantFactory):
            AUTO = True
            CONVERT_NAME = lambda x: chain_functions(
                x, Text.split_capitals_with_underscore, Text.lower
            )

        class SubA(A):
            pass

        sub_a = SubA()
        assert sub_a.name == "sub_a"


    def test_create_from_subclass(self):
        assert JsonConfigurationFileExtension is ConfigurationFileExtension.create(JsonConfigurationFileExtension, return_type=True)

    def test_create_from_subclass2(self):
        json_config_extension = JsonConfigurationFileExtension()
        assert json_config_extension is ConfigurationFileExtension.create(json_config_extension)

    def test_get_registrant(self):
        assert JsonConfigurationFileExtension is ConfigurationFileExtension.get_registrant(JsonConfigurationFileExtension)

    def test_register_func(self):

        class A(RegistrantFactory):
            @classmethod
            def create(cls, a):
                return super().create(registrant_name=type(a), a=a)

            def __init__(self, a):
                self.a = a

        @A.register_func((dict,))
        def create_a_from_type(a):
            return A(a)

        assert {} == A.create({}).a


from pytest import raises
from creationism.registration.factory import RegistrantNotRegisteredError
from creationism.registration.factory import RegistrantFactory
from creationism.registration.utils import Text, chain_functions


class TestErrors:
    def test_empty_register_error(self):
        with raises(RegistrantNotRegisteredError) as errors:

            class A(RegistrantFactory):
                AUTO = True
                CONVERT_NAME = lambda x: chain_functions(
                    x, Text.split_capitals_with_underscore, Text.lower
                )
                REPLACE=True
            A.create("sub_a")


    def test_registrant_not_registered_error(self):
        with raises(RegistrantNotRegisteredError) as errors:

            class A(RegistrantFactory):
                AUTO = True
                CONVERT_NAME = lambda x: chain_functions(
                    x, Text.split_capitals_with_underscore, Text.lower
                )
                REPLACE=True
            class SubA(A):
                pass

            A.create("sub_b")



from creationism.configuration.config import Configuration
from creationism.configuration.extensions import open_config
from pathlib import Path
from pytest import raises


class TestConfigError:
    def test_build_config_class(self):

        with raises(ValueError) as errors:
            config = {"a": [1, 2, 3]}
            config_dict = Configuration(
                name="1", config_value=config, modes=("default",)
            )

            config2 = {"a": {1, 2, 3}}
            config_dict.merge(config2)

    def test_build_config_class2(self):
        with raises(ValueError) as errors:
            config = {"a": "hello", "b": {"c": {1, 2, 3}}}
            config_dict = Configuration(
                name="1", config_value=config, modes=("default",)
            )

            config2 = {"b": [1, 2, 3]}
            config_dict.merge(config2)

    def test_build_config_not_found(self):
        with raises(FileNotFoundError) as errors:
            user_config = str(Path(__file__).parent / "testobject2.json")
            config_value = open_config(user_config, search_paths=("./",))
            configuration = Configuration(
                name="configuration", config_value=config_value, modes=("default",)
            )
            builds = configuration.build_instances()



from creationism.registration.factory import RegistrantFactory
from creationism.registration.utils import Text, chain_functions

class TestRegistrantFactory:

    def test_create_instance(self):
        class A(RegistrantFactory):
            AUTO = True
            CONVERT_NAME = lambda x: chain_functions(
                x, Text.split_capitals_with_underscore, Text.lower
            )
            REPLACE = True

        class SubA(A):
            pass

        suba = A.create('sub_a')
        assert isinstance(suba, SubA)

        
    def test_create_type(self):
        class A(RegistrantFactory):
            AUTO = True
            CONVERT_NAME = lambda x: chain_functions(
                x, Text.split_capitals_with_underscore, Text.lower
            )
            REPLACE = True

        class SubA(A):
            pass

        suba = A.create('sub_a',return_type=True)
        assert SubA is suba

        
    def test_create_static_instance(self):
        class A(RegistrantFactory):
            AUTO = True
            CONVERT_NAME = lambda x: chain_functions(
                x, Text.split_capitals_with_underscore, Text.lower
            )
            REPLACE = True
            STATIC=True
            

        class SubA(A):
            pass

        suba = A.create('sub_a')
        suba2 = A.create('sub_a')
        assert suba is suba2
    
    def test_create_non_static_instance(self):
        class A(RegistrantFactory):
            AUTO = True
            CONVERT_NAME = lambda x: chain_functions(
                x, Text.split_capitals_with_underscore, Text.lower
            )
            REPLACE = True
            STATIC=False

        class SubA(A):
            pass

        suba = A.create('sub_a')
        suba2 = A.create('sub_a')
        assert suba is not suba2



from creationism.registration.utils import Text, chain_functions


class TestTextUtils:
    def test_lower(self):
        assert Text.lower("A") == "a"

    def test_split_capitals_with_underscore(self):
        assert Text.split_capitals_with_underscore("aBCdEf") == "a_B_Cd_Ef"
        assert Text.split_capitals_with_underscore("BCdEF") == "B_Cd_E_F"


def test_chain_functions():
    test_chain_function = lambda x: chain_functions(
        x, Text.split_capitals_with_underscore, Text.lower
    )
    assert test_chain_function("aBCdEf") == "a_b_cd_ef"
